import logging as log
from src.classes import MolecularData, ForceFieldInfo

def parse_cgen(file_path):
    molecular_data = MolecularData()
    
    #selection indicates where to store data in ForceFieldInfo()
    current_selection = None

    #CGEN has some lines to skip that aren't parameters OR comments...
    #might need to add things to this list if you are having issues with reading files. 
    to_skip = ('*', '#', '!', ';', 'END ', 'RETURN', ' ', 'RESI')

    with open(file_path, 'r') as file:
        for line in file:
            line = line.strip()
            if not line or line.startswith(to_skip):
                continue
            elif not line.strip():
                current_selection = None

            ####----merged.itp----####
            if line.startswith('ATOM'):
                current_selection = 'ATOM'
            elif line.startswith('BOND '): #SINGULAR-merged.rtp
                current_selection = 'BOND'  #SINGULAR-merged.rtp
            elif line.startswith('IMPR '):
                current_selection = 'IMPR'
            
            ####----ffbonded.itp----####
            elif line.startswith('BONDS'): #PLURAL-ffbonded.itp
                current_selection = 'BONDS'  #PLURAL-ffbonded.itp
            elif line.startswith('ANGLES'):
                current_selection = 'ANGLES'
            elif line.startswith('DIHEDRALS'):
                current_selection = 'DIHEDRAL'
            elif line.startswith('IMPROPERS'):
                current_selection = 'IMPROPER'
            # Add more sections as needed

            #FOR SELECTIONS
            
            if current_selection == 'ATOM':
                atom_data = line.split()
                row = {
                    'i': atom_data[1],
                    'j': atom_data[2],
                    'charge': atom_data[3]
                }
                molecular_data.add_atom(row)


#HANDLES THE PLURAL AND SINGLUAR VERSIONS THAT CgenFF likes to use.
#similar treatment to 
            elif current_selection == 'BOND ' and len(line.split()) != 1: #avoids header
                bond_data = line.split()
                row = {
                'i': bond_data[1],
                'j': bond_data[2]
                }
                molecular_data.add_connectivity(row)
            elif current_selection == 'BONDS' and len(line.split()) != 1:
                connectivity_data = line.split()
                row = {
                'i': connectivity_data[0],
                'j': connectivity_data[1],
                'func': int(1),
                'kb': connectivity_data[2], ###UNITS spring constant
                'b0': connectivity_data[3]  ###UNITS A -> nm
                }
                molecular_data.add_bonds(row)



#angles: ;      i        j        k  func       theta0       ktheta      ub0       kub

            elif current_selection == 'ANGLES' and len(line.split()) != 1 :
                angle_data = line.split()
                # Only some entries have Urey-Bradley potentials. Otherwise, the entry for
                #   these columns is zero.
                if angle_data[6] == '!':
                    row = {
                        'i': angle_data[0],
                        'j': angle_data[1],
                        'k': angle_data[2],
                        'func': 5,
                        'ktheta': angle_data[3],
                        'theta0': angle_data[4],
                        'kub': float(0.00),
                        'ub0': float(0.00)
                    }
                else:
                        row = {
                        'i': angle_data[0],
                        'j': angle_data[1],
                        'k': angle_data[2],
                        'func': 5, 
                        'ktheta': angle_data[3],
                        'theta0': angle_data[4],
                        'kub': angle_data[5],
                        'ub0': angle_data[6]
                    }
                molecular_data.add_angles(row)
            
            elif current_selection == 'DIHEDRAL' and len(line.split()) != 1:
                dihedral_data = line.split()
                row = {
                    'i': dihedral_data[0],  #STANDARD DIHEDRALS HAVE FUNCTION 9
                    'j': dihedral_data[1],
                    'k': dihedral_data[2],
                    'l': dihedral_data[3],
                    'func': int(9),
                    'kphi': dihedral_data[4],
                    'multi': dihedral_data[5],
                    'phi0': dihedral_data[6]
                }
                molecular_data.add_dihedrals(row)


# ;   i       j       k       l       func    phi0 [deg]    kphi [kJ/mol]          mult


            elif current_selection == 'IMPROPER' and len(line.split()) != 1:
                
                improper_data = line.split()
                row = {
                    'i': improper_data[0],    #IMPROPERS HAVE FUNCTION 2
                    'j': improper_data[1],
                    'k': improper_data[2],
                    'l': improper_data[3],
                    'kphi': improper_data[4],
                    #'multi': improper_data[5], # there are NO multiplicity functions here
                    'func': str(2),
                    'phi0': improper_data[6]
                }
                
                molecular_data.add_impropers(row)

    return molecular_data

def parse_ff(ff_file): ##WIP
    selection = False
    ff_data = ForceFieldInfo()

    with open(ff_file, 'r') as file:
        for line in file:
            try:
                data = line.split()
            except IndexError:
                continue

            # Checks for comments
            if not line.strip() or line.startswith(('*', '#', ';', '!')):
                continue
            if line.startswith('[ '):
                selection = line.split('[')[-1].split(']')[0].strip().lower()
                continue

            # Adheres to standard CHARMM forcefield headers, but subject to change.
            if selection == 'bondtypes':
                try:
                    bond_data = line.split()
                    existing_data = {
                        'i': bond_data[0],
                        'j': bond_data[1]
                    }
                    ff_data.add_bonds(existing_data)
                except IndexError:
                    continue
            
            # Adheres to standard CHARMM forcefield headers, but subject to change.
            elif selection == 'angletypes':
                try:
                    angle_data = line.split()
                    existing_data = {
                        'i': angle_data[0],
                        'j': angle_data[1],
                        'k': angle_data[2]
                    }
                    ff_data.add_angles(existing_data)
                except IndexError:
                    continue

            # Adheres to standard CHARMM forcefield headers, but subject to change.
            # 'func' being 9 is the only thing that makes this an improper.  
            elif selection == 'dihedraltypes' and data[4] == '9': #ACTUAL DIHEDRALS
                try:
                    dihedral_data = line.split()
                    existing_data = {
                        'i': dihedral_data[0],
                        'j': dihedral_data[1],
                        'k': dihedral_data[2],
                        'l': dihedral_data[3],
                        'multi': dihedral_data[7] 
                    }
                    ff_data.add_dihedrals(existing_data)
                except IndexError:
                    continue

            # Adheres to standard CHARMM forcefield headers, but subject to change.
            elif selection == 'dihedraltypes' and data[4] == '2': #IMPROPERS
                try:
                    improper_data = line.split()
                    existing_data = {
                        'i': improper_data[0],
                        'j': improper_data[1],
                        'k': improper_data[2],
                        'l': improper_data[3],
                        'func':improper_data[4]
                    }
                    ff_data.add_impropers(existing_data)
                except IndexError:
                    continue

    return ff_data


import logging as log
import pandas as pd
import re
import os
import config


# import user flags stored in config file
input_file_CHARMM = config.input_file_CHARMM
input_file_CGEN = config.input_file_CGEN
output_file = config.output_file
unit_convert = config.convert_to_kj


class MolecularData:
    def __init__(self):
        
        #broken down by file type for future functionality
        self.data = {
            'merged.rtp': {
                'atom': [],
                'connectivity': []
            },
            'ffbonded.itp': {
                'bonds': [],
                'angles': [],
                'dihedrals': [],
                'impropers': []
            }
        }

#store data in the class and empty arrays above
    def add_atom(self, atom_data):
        self.data['merged.rtp']['atom'].append(atom_data)

    def add_connectivity(self, connectivity_data):
        self.data['merged.rtp']['connectivity'].append(connectivity_data)

    def add_bonds(self, bond_data):
        self.data['ffbonded.itp']['bonds'].append(bond_data)
    def add_angles(self, angle_data):
        self.data['ffbonded.itp']['angles'].append(angle_data)

    def add_dihedrals(self, dihedral_data):
        self.data['ffbonded.itp']['dihedrals'].append(dihedral_data)

    def add_impropers(self, improper_data):
        self.data['ffbonded.itp']['impropers'].append(improper_data)

#retrieve data from the class

    def get_atoms(self):
        #not used yet; the issue of unique entries is not present in merged.rtp
        return pd.DataFrame(self.data['merged.rtp']['atom'])

    def get_connectivity(self):
        #not used yet; the issue of unique entries is not present in merged.rtp
        return pd.DataFrame(self.data['merged.rtp']['connectivity'])

    def get_bonds(self):
        df = pd.DataFrame(self.data['ffbonded.itp']['bonds'])
        df['b0'] = pd.to_numeric(df['b0'], errors='coerce')
        df['kb'] = pd.to_numeric(df['kb'], errors='coerce')

        #CGEN uses kcal by default, but GROAMCS uses kJ
        if unit_convert:
            b0_factor = 0.1 #Å -> nm
            kb_factor = 836.8 #kcal/molÅ^-2 -> kJ/mol*nm^-2 #CORRECT FOR FUNC_1
            df['b0'] *= b0_factor
            df['kb'] *= kb_factor

            unit_dict = {
                'b0': 'nm',
                'kb': 'kJ/mol*nm^-2'}
        
        else:
            unit_dict = {
                'b0': 'Å',
                'kb': 'kcal/mol*Å^2'
            }

        # Rename the unit columns to distinguish from the original columns
        units_df = pd.DataFrame({key: [value] for key, value in unit_dict.items()}) # (1, 4)
        units_df = pd.concat([units_df] * len(df), ignore_index=True) # (len(parsed_entries), 4) each column is the same value
        units_df = units_df.add_suffix('_unit')
        
        df_with_units = pd.concat([df, units_df], axis=1)# (len(parsed_entries), 11)

        # df containing all entries from cgen file with units in each row of *_unit keys
        return df_with_units

#angles: ;      i        j        k  func       theta0       ktheta      ub0       kub

    def get_angles(self):
        df = pd.DataFrame(self.data['ffbonded.itp']['angles'])
        df[['ktheta', 'theta0', 'kub', 'ub0']] = df[['ktheta', 'theta0', 'kub', 'ub0']].apply(pd.to_numeric, errors='coerce')

        #CGEN uses kcal by default, but GROAMCS uses kJ
        if unit_convert:
            theta_factor = 1 #deg -> deg
            ktheta_factor = 8.368 #kcal -> kJ * func_5_factor
            ub0_factor = 0.1 #Å -> nm
            kub_factor = 836.8 #kcal/molÅ^-2 -> kJ/mol*nm^-2 #CORRECT FOR FUNC_5
            df['theta0'] *= theta_factor #for redundancy, units are same (deg)
            df['ktheta'] *= ktheta_factor
            df['kub'] *= kub_factor
            df['ub0'] *= ub0_factor

            #UNITS
            unit_dict = {
                'ktheta': 'kJ/mol',
                'theta0': 'deg',
                'kub': 'kJ/mol*nm^-2',
                'ub0': 'nm'} 
        else:
            #UNITS
            unit_dict = {
                'ktheta': 'kcal/mol',
                'theta0': 'deg',
                'kub': 'kcal/mol*Å^-2',
                'ub0': 'Å'} 

        # Rename the unit columns to distinguish from the original columns
        units_df = pd.DataFrame({key: [value] for key, value in unit_dict.items()}) # (1, 4)
        units_df = pd.concat([units_df] * len(df), ignore_index=True) # (len(parsed_entries), 4) each column is the same value
        units_df = units_df.add_suffix('_unit')
        df_with_units = pd.concat([df, units_df], axis=1)# (len(parsed_entries), 11)

        # df containing all entries from cgen file with units in each row of *_unit keys
        return df_with_units

    def get_dihedrals(self):
        df = pd.DataFrame(self.data['ffbonded.itp']['dihedrals'])
        df[['kphi', 'phi0', ]] = df[['kphi', 'phi0']].apply(pd.to_numeric, errors='coerce')

        #CGEN uses kcal by default, but GROAMCS uses kJ
        if unit_convert:
            phi_0_factor = 1 #deg -> deg
            kphi_factor = 4.184 #kcal -> kJ

            df['kphi'] *= kphi_factor #TO BE MULTIPLIED BY MULTI FOR THIS FUNCTION
            df['phi0'] *= phi_0_factor

            unit_dict = {
                'kphi': 'kJ/mol',
                'phi0': 'deg'}
        else:
            unit_dict = {
                'kphi': 'kcal/mol',
                'phi0': 'deg'}

        # Rename the unit columns to distinguish from the original columns
        units_df = pd.DataFrame({key: [value] for key, value in unit_dict.items()}) # (1, 4)
        units_df = units_df.add_suffix('_unit')
        units_df = pd.concat([units_df] * len(df), ignore_index=True) # (len(parsed_entries), 4) each column is the same value
        
        df_with_units = pd.concat([df, units_df], axis=1)# (len(parsed_entries), 11)

        # df containing all entries from cgen file with units in each row of *_unit keys
        return df_with_units


    def get_impropers(self):
        df = pd.DataFrame(self.data['ffbonded.itp']['impropers'])
        
        try: #often empty
            df[['kphi', 'phi0']] = df[['kphi', 'phi0']].apply(pd.to_numeric, errors='coerce')

        except KeyError as e:
            print(f'Key Error! {e}', 'is the parameter field entry?')
        
        #CGEN uses kcal by default, but GROAMCS uses kJ
        if unit_convert:
            keta_factor = 8.368 #kcal/mol -> kJ/mol *2
            eta0_factor = 1 #deg -> deg 
            df['kphi'] *= keta_factor
            df['phi0'] *= eta0_factor

            unit_dict = {
                'kphi': 'kJ/mol',
                'phi0': 'deg'}
        else:
            unit_dict = {
                'kphi': 'kcal/mol',
                'phi0': 'deg'}

        units_df = pd.DataFrame({key: [value] for key, value in unit_dict.items()}) # (1, 2)
        units_df = pd.concat([units_df] * len(df), ignore_index=True) # (len(parsed_entries), 2) each column is the same value
        units_df = units_df.add_suffix('_unit')        

        df_with_units = pd.concat([df, units_df], axis=1)# (len(parsed_entries), 11)

        # df containing all entries from cgen file with units in each row of *_unit keys
        return df_with_units

    #HARDCODED headers that are standard for modern CHARMM forcefields
    def bond_header(self):
        header = '[ bondtypes ]' +'\n'+'; i        j          func      b0[b0_unit]         kb[kb_unit]'
        return str(header)

    def angle_header(self):
        header = '[ angletypes ]' + '\n'+'; i        j        k           func  theta0[theta0_unit]  ktheta[ktheta_unit]  ub0[ub0_unit]    kub[kub_unit]'
        return str(header)

    def dihedral_header(self):
        header = '[ dihedraltypes ]'+'\n'+'; i       j       k       l         func   phi0[phi0_unit]   kphi[kphi_unit] mult'
        return str(header)

    def improper_header(self):
        header = '[ dihedraltypes ] ;IMPROPERS'+'\n'+'; i       j       k       l         func   phi0[phi0_unit]   kphi[kphi_unit]'
        return(str(header))


#nearly identical class to MolecularData, but less functionality

class ForceFieldInfo:

    def __init__(self):
        self.data = {
            'merged.rtp': {
                'atom': [],
                'connectivity': []
            },
            'ffbonded.itp': {
                'bonds': [],
                'angles': [],
                'dihedrals': [],
                'impropers': []
            }
        }

    def add_atom(self, atom_data):
        self.data['merged.rtp']['atom'].append(atom_data)

    def add_connectivity(self, connectivity_data):
        self.data['merged.rtp']['connectivity'].append(connectivity_data)

    def add_bonds(self, bond_data):
        self.data['ffbonded.itp']['bonds'].append(bond_data)

    def add_angles(self, angle_data):
        self.data['ffbonded.itp']['angles'].append(angle_data)

    def add_dihedrals(self, dihedral_data):
        self.data['ffbonded.itp']['dihedrals'].append(dihedral_data)

    def add_impropers(self, improper_data):
        self.data['ffbonded.itp']['impropers'].append(improper_data)

    def get_atoms(self):
        return pd.DataFrame(self.data['merged.rtp']['atom'])

    def get_connectivity(self):
        return pd.DataFrame(self.data['merged.rtp']['connectivity'])

    def get_bonds(self):
        return pd.DataFrame(self.data['ffbonded.itp']['bonds'])

    def get_angles(self):
        return pd.DataFrame(self.data['ffbonded.itp']['angles'])

    def get_dihedrals(self):
        return pd.DataFrame(self.data['ffbonded.itp']['dihedrals'])

    def get_impropers(self):
        return pd.DataFrame(self.data['ffbonded.itp']['impropers'])




import logging as log
import pandas as pd
from src.parse_files import parse_cgen, parse_ff
import os
import click
import sys
import config

# How to loop through the dictionary and where the keys are addressable.
def iterate_nested_dict(nested_dict, parent_key=''):
    for key, value in nested_dict.items():
        full_key = f"{parent_key}.{key}" if parent_key else key
        if isinstance(value, dict):
            iterate_nested_dict(value, full_key)
        else:
            print(f"{full_key}")


##EXAMPLE
# print(cgen.data['ffbonded.itp']['bonds'][0]) #DOES NOT CONTAIN UNITS 
# print(cgen.get_angles().keys()) #CONTAINS UNITS 


# WIP, checking for backwards entries is off by default. 
def create_entries_set(df, index_columns):
    entries_set = set()
    for row in df[index_columns].values:
        direct_entry = tuple(row)
        reversed_entry = tuple(row[::-1])
        entries_set.add(direct_entry)
        entries_set.add(reversed_entry)
    return entries_set


def get_uniques(ff, cgen, index_columns, param):

    try:
        if config.reverse_entries: #WIP
            print('WIP--not checking for backwards entries')
            pass
        ##JUMBLES THE ORDER???
            ff_entries = create_entries_set(ff, index_columns) 
            cgen_entries = create_entries_set(cgen, index_columns)
            print('checking reverse entries...')
    
            # # screen for unique entries into df
            unique_entries = [entry for entry in cgen_entries if entry not in ff_entries]
            unique_entries_df = pd.DataFrame(unique_entries, columns=index_columns)
    
            # this merge checks for duplicates in 4 columns and merges them with EXISTING
            # parameters found in CGEN file
            unique_entries_df = unique_entries_df.merge(cgen, on=index_columns, how='left')
    
            columns_to_drop = ff.columns.difference(index_columns)
            unique_entries_df = unique_entries_df.drop(columns=columns_to_drop, errors='ignore')
            unique_entries_df.reset_index(drop=True, inplace=True)
            unique_entries_df = unique_entries_df.dropna()
    
            unique_entries_df.reset_index(drop=True, inplace=True)  
            return unique_entries_df
        
        # Traditional matches
        # merging two dfs
        merged = cgen.merge(ff, on=index_columns, how='left', indicator=True)
        
        # merge both arrays
        unique_entries = merged[merged['_merge'] == 'left_only']

        # Drop columns from ff_df that are not in index_columns
        columns_to_drop = ff.columns.difference(index_columns)
        columns_to_drop = [col for col in columns_to_drop if col in unique_entries.columns]
        unique_entries_df = unique_entries.drop(columns=columns_to_drop)
        unique_entries_df = unique_entries.drop(columns=['_merge', 'mult_y'], errors='ignore')  # Drop unnecessary columns

        # Remove columns and reset index
        unique_entries_df.reset_index(drop=True, inplace=True)
        return unique_entries_df
        
    except AttributeError as e:
        print(f'AttributeError: {e}')
        print(f'cannot merge empty array, possibly there are no entires for {param}?')
    except KeyError as e:
        if len(ff.columns) == 0:
            print(f'KeyError: {e}')
            print(f'length of merged array for {param} is 0, no unique entries detected?')
        else:
            print(f'KeyError: {e}')
            print(f'ff_df columns: {ff.columns}')
            print(f'index_columns: {index_columns}')


# print(unique_bonds['kb_unit'].iloc[0])

def format_string(df, unformatted_string):

    # Screen for type of parameter based on headers from incoming dataframe 
    # There may be a more flexible way to do this
    if all(col in df.columns for col in ['i', 'j', 'func', 'kb', 'b0']): #bonds
        format_template = "{:>8}  {:>8}  {:>8}   {:>8.5f}        {:>8.2f}" 
        
        #REORDER AND RECAST
        split_line = unformatted_string.split()
        func = int(split_line[2])
        kb = float(split_line[3])
        b0 = float(split_line[4])
        
        reordered_line = split_line[:2] + [func, b0, kb]
        line  = format_template.format(*reordered_line)

    elif all(col in df.columns for col in ['i', 'j', 'k', 'func', 'ktheta', 'theta0','kub', 'ub0']): #angles
        format_template = "{:>8}{:>8}{:>8}   {:>8}    {:>8.3f}   {:>8.5f}  {:>6.4f}  {:>8.4f}"
        
        #REORDER AND RECAST
        split_line = unformatted_string.split() 
        ktheta = float(split_line[4])
        theta0 = float(split_line[5])
        r0 = float(split_line[7])
        kub = float(split_line[6])

        reordered_line = split_line[:4] + [theta0, ktheta, r0, kub]
        line = format_template.format(*reordered_line)

    elif all(col in df.columns for col in ['i', 'j', 'k', 'l', 'func', 'kphi', 'multi', 'phi0']): #dihedrals
        format_template = "{:>8}{:>8}{:>8}{:>8}{:>8}    {:>8.4f}   {:>8.5f}     {:>1}"

        #REORDER AND RECAST
        split_line = unformatted_string.split()
        kphi = float(split_line[5])
        multi = int(split_line[6])
        phi0 = float(split_line[7])

        reordered_line = split_line[:5] + [phi0, kphi, multi]
        line = format_template.format(*reordered_line)

    elif all(col in df.columns for col in ['i', 'j', 'k', 'l', 'func', 'kphi', 'phi0']): #impropers
        format_template = "{:>8}{:>8}{:>8}{:>8}{:>8}    {:>8.4f}   {:>8.5f}"

        #REORDER AND RECAST
        split_line = unformatted_string.split()
        kphi = float(split_line[4])
        func = split_line[5]
        phi0 = float(split_line[6])

        reordered_line = split_line[:4] + [func, phi0, kphi]
        line = format_template.format(*reordered_line)
    
    return line


def update_charmm(df, outfile, header, param):

    # Create a mapping between unit columns and their respective units
    # Necessary when reordering of the columns in format_string()
    unit_mapping = {}
    unit_cols = [col for col in df.columns if '_unit' in col]
    try:
        for col in unit_cols:
            unit_mapping[col.replace('_unit', '')] = df[col].iloc[0] 
    
    # to circumvent an empty array from get_uniques() which doesn't have df[col].iloc[0]
    except IndexError as e:
        unit_mapping = {}
        print(e, ' *from update_charmm() from get_uniques.py*')
        print(f' - It is likely that no unique {param} were found in --itp input')
    
    # This redefine is crucial here
    formatted_header = header

    #match each '*_unit' key to its matching '*' key
    for key, unit in unit_mapping.items():
        formatted_header = formatted_header.replace(f'[{key}_unit]', f'[{unit}]')

    file_exists = os.path.exists(outfile)
    with open(outfile, 'a' if file_exists else 'w') as f:
        f.write(formatted_header + '\n')

        if df is not None:
            df = df.fillna(float(0.0))
            header_columns = [col for col in df.columns if '_unit' not in col]
            for index, row in df.iterrows():
                unformatted_line = '   '.join(str(row[col]) if not isinstance(row[col], list) else ' '.join(map(str, row[col])) for col in header_columns)

                formatted_line = format_string(df, unformatted_line)
                f.write(formatted_line + '\n')






from setuptools import setup, find_namespace_packages
import pathlib

here = pathlib.Path(__file__).parent.resolve()

# Get the long description from the README file
long_description = (here / "README.md").read_text(encoding="utf-8")

setup(
    name="cgen2gmx",  
    version="1.0",  # Required
    long_description=long_description, 
    long_description_content_type="text/markdown", 
    url="https://github.com/chrispy67/cgen2gmx",
    author="Christian Phillips", 
    author_email="christian_phillips1@msn.com",
    classifiers=[
        "Development Status :: 1 - Planning",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
    ],
    keywords="molecular dynamics, charmm, forcefield, computational chemistry",
    packages=find_namespace_packages(),

    python_requires="<4",
    extras_require={ 
        "test": ["coverage", 'pytest'],
    },

    package_data={},
    data_files=None, 
    entry_points={},
)